/*
 * Copyright (c) 2000 Paul Herman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef TCPPROF_TCPPROF_H
#define TCPPROF_TCPPROF_H

#ifdef HAVE_CONFIG_H
# include "config.h"
#else  /* Hmmm, try to wing it */
# ifdef __linux__
# define _BSD_SOURCE
# endif
#endif /* HAVE_CONFIG_H */

#include <stdio.h>		/* printf() */
#include <unistd.h>		/* fork(), execlp() */
#include <stdlib.h>		/* free(), malloc() */
#include <string.h>		/* bzero() */
#include <math.h>		/* sqrt() */
#include <ctype.h>		/* tolower */

#include <sys/time.h>		/* struct timeval */
#include <sys/wait.h>		/* waitpid() */

#include <sys/socket.h>		/* needed by INET stuff (sockaddr) */
#include <netinet/in.h>		/* needed by INET stuff (in_addr) */
#include <netinet/in_systm.h>	/* needed by INET stuff (n_long) */

#include <net/ethernet.h>	/* struct ether_header */
#include <net/if_arp.h>		/* struct arp_hdr */
#include <netinet/ip.h>		/* struct ip */
#include <netinet/tcp.h>	/* struct tcphdr */
#include <netinet/udp.h>	/* struct udphdr */
#include <netinet/ip_icmp.h>	/* struct icmp */

#include <arpa/inet.h>		/* inet_ntoa() */

#include <netdb.h>		/* getprotobynumber */
#include <err.h>

#include <pcap.h>		/* pcap_*() */

#define BUF_SIZ		4096		/* General string buffer size */
#define PAK_SIZ		32768
#define Double		double		/* FP accuracy */

	/* Of the "LINK" statistic, types of link layers */
enum link_type	{ NONE, ETHERNET_IP, ETHERNET_ARP };

typedef struct stat_unit {
	u_int		key;
	u_quad_t	bytes;
	u_quad_t	num_packets;
} stat_unit;


	/* Types of statistics to measure */
#define TYPE_ALL		0x01
#define TYPE_LINK		0x02
#define TYPE_IP_PROTO		0x04
#define TYPE_PORT		0x08
#define TYPE_HOST		0x10
#define TYPE_NET		0x20

	/* in get_tcpdump_data, flags on how to operate */
#define GET_TCPD_COUNT_LINKSIZE		0x01
#define GET_TCPD_DO_LIVE		0x02
#define GET_TCPD_DO_LIVE_PROMISC	0x04

typedef struct stat_root {
	u_int		type;
	u_int		count;
	struct stat_unit **hash;
} stat_root;


typedef struct ip_packet {
	struct ip	hdr;
	union {
		struct tcphdr	tcphdr;
		struct udphdr	udphdr;
		struct icmp	icmp;
	} body;
} ip_packet;

typedef struct packet_data {
	struct timeval		timestamp;
	u_int			packet_len;
	u_char			link_type;
	struct ether_header	ether;
	union {
		struct arphdr		arp;
		struct ip_packet	ip;
		u_char			raw[PAK_SIZ];
	} data;
} packet_data;

/************** Prototypes **************/

/* main.c protos */
void process_file(char *fname, u_int stats_types);
int Usage(int r, char *prog);
void error(char *s);

/* process.c protos */
void process_tcpdump_header(char *l, packet_data *p, int count_ether_size);
void get_tcpdump_data(char *fname, char *filter, int flags,
        Double capture_seconds, void (*hook)(packet_data *, void **),
	void **args);

/* print_packet.c protos */
void print_packet(packet_data *p);
void print_arp_header(struct arphdr *arp);
void print_ip_header(struct ip *ip);
void print_tcp_header(struct tcphdr *tcp);
void print_udp_header(struct udphdr *udp);
void print_icmp_header(struct icmp *icmp);

/* stats.c protos */
void add_packet_to_root(stat_root *roots, packet_data *pd, u_int types);
void free_packets(stat_root *roots, int tc);
void show_results(stat_root *roots, u_int types);
 
/* utils.c protos */
char *my_basename(char *p);
void timeval_diff(struct timeval *d, struct timeval *t2, struct timeval *t1);
void timeval_avg(struct timeval *d, struct timeval *t2, struct timeval *t1);
int timeval_lessthan(struct timeval *t1, struct timeval *t2);
int timeval_lessthaneq(struct timeval *t1, struct timeval *t2);
void strtoargv(char **argv, char *s, int *idx);
char *getfirstchar(char *s);
char *my_strsep(char **p, char d);
u_char count_1bits(u_int x);
char *my_get_port(u_short port, int n_only);
char *my_get_proto(u_short proto, int n_only);
int key_compare(const void *h1, const void *h2);
int bytes_compare(const void *h1, const void *h2);
char ether_addr_is_zero(u_char addr[ETHER_ADDR_LEN]);
struct ether_addr *my_ether_aton(char *c);
char *my_safe_strcpy(char **dst, const char *src);

/* quick_sqrt.c protos */
double quick_sqrt(double x);

#endif

/**************** END *****************/
