/*
 * Copyright (c) 2000 Paul Herman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "tcpprof.h"

void print_arp_header(struct arphdr *arp);
void print_ip_header(struct ip *ip);
void print_tcp_header(struct tcphdr *tcp);
void print_udp_header(struct udphdr *udp);
void print_icmp_header(struct icmp *icmp);

void print_packet(packet_data *p) {
	struct ip	*ip;
	struct tcphdr	*tcp;
	struct udphdr	*udp;
	struct icmp	*icmp;

	printf("PACKET SIZE: %d\n", p->packet_len);
	if (ntohs(p->ether.ether_type) == ETHERTYPE_ARP) {
		print_arp_header(&(p->data.arp));
		printf("---------------------------\n");
		return;
		}

	/* OK, it's an IP Packet */

	ip = &(p->data.ip.hdr);
	tcp= &(p->data.ip.body.tcphdr);
	udp= &(p->data.ip.body.udphdr);
	icmp= &(p->data.ip.body.icmp);

	print_ip_header(ip);
	switch (ip->ip_p) {
		case IPPROTO_TCP:
			print_tcp_header(tcp); 	break;
		case IPPROTO_UDP:
			print_udp_header(udp); 	break;
		case IPPROTO_ICMP:
			print_icmp_header(icmp); break;
		default:
			printf("UNKNOWN PROT!\n");break;
		}
	printf("---------------------------\n");
}

void print_arp_header(struct arphdr *arp) {
		printf("ARP PACKET:\n");
		printf("\thrd=%d pro=%d hln=%d plen=%d op=%d",
			htons(arp->ar_hrd),
			htons(arp->ar_pro),
			arp->ar_hln,
			arp->ar_pln,
			htons(arp->ar_op)
			);
		printf("\n");
}

void print_ip_header(struct ip *ip) {
	struct protoent *ip_prot;
#if 0
	u_char *p, i;
	p = (char *)ip;
	for (i=0; i<sizeof(struct ip); i++)
		printf("%.2X ", *p++);
	printf("\n");
#endif
	printf("IP HEADER:\n");
	ip_prot = getprotobynumber(ip->ip_p);
	printf("\tver=%d hlen=%d TOS=%d len=%d ID=0x%.2X",
#ifdef _IP_VHL
		ip->ip_vhl >> 4, (ip->ip_vhl & 0x0f) << 2,
#else
		ip->ip_v, ip->ip_hl<<2,
#endif
		ip->ip_tos,
		htons(ip->ip_len),
		htons(ip->ip_id)
		);
	printf("\n\tFRAG=0x%.2x TTL=%d Proto=%s cksum=0x%.2X\n",
		ip->ip_off,
		ip->ip_ttl,
		ip_prot->p_name,
		htons(ip->ip_sum)
		);

	printf("\t%s", inet_ntoa(ip->ip_src) );
	printf(" -> %s\n", inet_ntoa(ip->ip_dst) );

}

void print_tcp_header(struct tcphdr *tcp) {
#if 0
	u_char *p, i;
	p = (char *)tcp;
	for (i=0; i<sizeof(struct tcphdr); i++)
		printf("%.2X ", *p++);
	printf("\n");
#endif

	printf("\tTCP HEADER:\n");
	printf("\t\tsport=%d dport=%d seq=0x%lX ack=0x%lX doff=0x%.2X\n",
		htons(tcp->th_sport),
		htons(tcp->th_dport),
		(unsigned long int)htonl(tcp->th_seq),
		(unsigned long int)htonl(tcp->th_ack),
		tcp->th_off
		);
	printf("\t\tflags=(0x%X)", tcp->th_flags);
	if (tcp->th_flags & TH_FIN)	printf(",FIN");
	if (tcp->th_flags & TH_SYN)	printf(",SYN");
	if (tcp->th_flags & TH_RST)	printf(",RST");
	if (tcp->th_flags & TH_PUSH)	printf(",PUSH");
	if (tcp->th_flags & TH_ACK)	printf(",ACK");
	if (tcp->th_flags & TH_URG)	printf(",URG");
	printf(" win=0x%.4X cksm=0x%.4X urp=0x%.4X",
		htons(tcp->th_win),
		htons(tcp->th_sum),
		htons(tcp->th_urp)
		);
	printf("\n");
}

void print_udp_header(struct udphdr *udp) {
#if 0
	u_char *p, i;
	p = (char *)udp;
	for (i=0; i<sizeof(struct udphdr); i++)
		printf("%.2X ", *p++);
	printf("\n");
#endif

	printf("\tUDP HEADER:\n");
	printf("\t\tsport=%d dport=%d len=%d cksum=0x%.2X\n",
		htons(udp->uh_sport),
		htons(udp->uh_dport),
		htons(udp->uh_ulen),
		htons(udp->uh_sum)
		);
}

void print_icmp_header(struct icmp *icmp) {
#if 0
	u_char *p, i;
	p = (char *)icmp;
	for (i=0; i<sizeof(struct icmp); i++)
		printf("%.2X ", *p++);
	printf("\n");
#endif

	printf("\tICMP HEADER:\n");
	printf("\t\ttype=0x%X code=0x%X cksum=0x%.4X\n",
		icmp->icmp_type,
		icmp->icmp_code,
		htons(icmp->icmp_cksum)
		);
}
