/*
 * Copyright (c) 2000 Paul Herman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "tcpprof.h"

char	numbers_only		= 0;
char	filterexpr[BUF_SIZ]	= "";
char	registered_only		= 0;
int	ports_to_show		= IPPORT_RESERVED;

#define USAGE 							\
"Usage: %s [-?hn] [-p port] [-s letters] [-f filter expr] filename
	-?, -h 		- display this help
	-f filter expr	- tcpdump(1) filter expression
	-n		- don't resolve network numbers
	-p port		- ignore port numbers >= port
	-r		- display only registered ports
	-s letters	- collect stats
		a		- collect ALL stats
		l		- collect link stats
		i		- collect ip stats
		p		- collect port stats
		h		- collect host stats
		n		- collect network stats

Example: %s -s ph data.dump
	Displays port and host statistics
"

void my_hook(packet_data *pd, void **args) {
	stat_root	*sr;
	u_int		*types;

	sr = (stat_root *) *args++; types = (u_int *) *args++;
	add_packet_to_root(sr, pd, *types);
}

/*
 * process_file() takes the output of tcpdump, saves packets, and creates
 * statistics
 */
void process_file(char *fname, u_int s_types) {
	int 		i, j, tc;
	stat_root	*stat_r;
	void		*argv[2];

	tc = count_1bits(s_types);
	if (tc == 0) return;
	stat_r = (stat_root *) malloc(tc*sizeof(stat_root));

	/* Init stat_r */
	for (i=0,j=0; i<tc && j<32; j++) {
		if ( (s_types>>j) & 0x1) {
			stat_r[i].type = 1<<j;
			stat_r[i].hash = NULL;
			i++;
			}
		}
	argv[0] = (void *) stat_r;
	argv[1] = (void *) &s_types;
	get_tcpdump_data(fname, filterexpr, GET_TCPD_COUNT_ETHERSIZE,
		-1.0, my_hook, argv);
	show_results(stat_r, s_types);

	free_packets(stat_r, tc);
	free(stat_r);
}

u_int parse_stats_types(char *in) {
	u_int ret = 0;
	while (*in) {
		if (*in == 'l' || *in == 'a' ) ret |= TYPE_LINK;
		if (*in == 'i' || *in == 'a' ) ret |= TYPE_IP_PROTO;
		if (*in == 'p' || *in == 'a' ) ret |= TYPE_PORT;
		if (*in == 'h' || *in == 'a' ) ret |= TYPE_HOST;
		if (*in == 'n' || *in == 'a' ) ret |= TYPE_NET;
		in++;
		}
	return ret;
}

int Usage(int r, char *prog) {
	fprintf(stderr, USAGE, my_basename(prog), prog);
	return r;
}

void error(char *s) {
	fprintf(stderr, "error: %s\n", s);
	Usage(-1, "tcpstat"); exit(-1);
}

int main(int argc, char **argv) {
	char	ch, *filename;
	u_int	stats_types = TYPE_ALL;

	while ( (ch = getopt(argc, argv, "h?np:rf:s:")) != -1) {
		switch(ch) {
			case 'h':
			case '?':
				return Usage(1, argv[0]);
				break;
			case 'f':
				strncpy(filterexpr, optarg, BUF_SIZ);
				break;
			case 'n':
				numbers_only = 1;
				break;
			case 'p':
				ports_to_show = (int) strtol(optarg, NULL, 10);
				registered_only = 1;
				break;
			case 'r':
				registered_only = 1;
				break;
			case 's':
				stats_types |= parse_stats_types(optarg);
				break;
			default:
				break;
			}				
		}

	if ((argc - optind) < 1) error("must specify filename");
	filename = argv[optind];

	process_file(filename, stats_types);

	return 0;
}
