/*
 * Copyright (c) 2000 Paul Herman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "tcpprof.h"

extern char numbers_only;
extern char sort_by_bytes;
extern char registered_only;
extern int ports_to_show;

const char *type_str[] = { "Total", "Link", "IP", "Port", "Host", "Network" };

void print_line(char *label, u_quad_t num_p, u_quad_t bytes, Double pcnt) {
	char str[1024];
	sprintf(str, "%s", label);
	printf("\t%-20s%-15qu%-15qu%-5.4f %%\n", str, num_p, bytes, pcnt);
}

u_int get_data_from_packet(u_int t, packet_data *p, char source) {
	u_int	ipa, r;
	u_short et;
	struct ether_header *eh;
	struct ip_packet *ipp;
	struct tcphdr *th;
	struct udphdr *uh;

	eh =  &(p->ether);
	ipp = &(p->data.ip);
	th = &(ipp->body.tcphdr);
	uh = &(ipp->body.udphdr);

		/* Keep track of loop back data */
	et = ntohs(eh->ether_type);
	if (et == ETHERTYPE_IP) {
		ipa = (source)?
			*(u_int*)&(ipp->hdr.ip_src) :
			*(u_int*)&(ipp->hdr.ip_dst);
		}
	switch (t) {
		case TYPE_ALL:
			return 2;
			break;
		case TYPE_LINK:
			return p->link_type + 1;
			break;
		case TYPE_IP_PROTO:
			if (et == ETHERTYPE_IP)
				return ipp->hdr.ip_p + 1;
			break;
		case TYPE_PORT:
			if (et != ETHERTYPE_IP) break;
			if (ipp->hdr.ip_p == IPPROTO_TCP)
				r = (source)?
					htons(th->th_sport) :
					htons(th->th_dport);
			else if (ipp->hdr.ip_p == IPPROTO_UDP)
				r = (source)?
					htons(uh->uh_sport) :
				 	htons(uh->uh_dport);
			else r = 0;
			if (r >= ports_to_show  && registered_only) r = 0;
			return r;
			break;
		case TYPE_HOST:	/* XXX: Assume 32 bit IP Addresses */
			if (et != ETHERTYPE_IP) break;
			return (source)?
				*(u_int*)&(ipp->hdr.ip_src) :
				*(u_int*)&(ipp->hdr.ip_dst);
			break;
		case TYPE_NET:
			if (et != ETHERTYPE_IP) break;
			return (source)?
				inet_netof(ipp->hdr.ip_src) :
				inet_netof(ipp->hdr.ip_dst);
			break;
		default: break;
		}
	return 0;
}

int mkey_compare(const void *h1, const void *h2) {
        const stat_unit *aa, *bb;
	
        aa = (const stat_unit *)h1; bb = (const stat_unit *)h2;
	return ((*aa).key < (*bb).key)? -1 : ((*aa).key > (*bb).key)? 1 : 0;
}


/* XXX: Can this be done with bsearch(3) ? */
stat_unit *find_entry(u_int type, size_t count, stat_unit **hash,
		u_int key) {
	size_t a;

	if (count == 0) return NULL;
#if 1
	for (a=0; a<count; a++)
		if (hash[a]->key == key) return hash[a];
	return NULL;
#else
		/* XXX: This needs to be fixed */
	stat_unit key_stat;
	key_stat.key = key;
	return (stat_unit *) bsearch((void *) &key_stat, hash[0],
			count, sizeof(void *), mkey_compare);
#endif
}

void free_packets(stat_root *roots, int tc) {
	int i;
	u_int j;
	stat_unit **p;
	for (i=0; i<tc; i++)
		for (j=0, p=roots[i].hash; j<roots[i].count; j++)
			free(p[j]);
}

void add_packet(u_int type, u_int *count, stat_unit ***hash,
		packet_data *p, u_int k) {
	stat_unit *u, **array;

	array = *hash;
	u = find_entry(type, *count, array, k);
	if (u != NULL) {
		u->bytes += p->packet_len;
		u->num_packets++;
			/* XXX: add timestamp entries */
/*		printf("Added %u bytes\n", p->packet_len); */
		}
	else {
		/* else add data at then end, and then sort hash */
		*count += 1;
		array = realloc(array, (*count) * sizeof(stat_unit) );
		u = (stat_unit *) malloc(sizeof(stat_unit));
		array[(*count)-1] = u;
		u->key = k;
		u->bytes = p->packet_len;
		u->num_packets = 1;
			/* XXX: add timestamp entries */
/*		printf("Added %u bytes\n", p->packet_len); */

		/* Sort data */
/*		qsort(&(array[0]), *count, sizeof(void *), key_compare); */
		*hash = array;
		}
}

void compact_type(u_int *count, stat_unit ***hash) {
	stat_unit **array;

	if (*count < 2) return;
	array = *hash;
}

void show_totals(u_int type, u_int *count, stat_unit ***hash,
		u_quad_t tb) {
	char str[256];
	u_int	i;
	Double pcnt;
	u_quad_t l, b;
	stat_unit *u;
	struct in_addr	*addr;

	b = 0;
	u = NULL;
	switch (type) {
		case TYPE_ALL:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];	b += u->bytes;
			}
		pcnt = 100.0 * (Double)b/(Double)tb;
		print_line( "Total", u->num_packets, b, pcnt);
		break;
		case TYPE_LINK:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];
			b = u->bytes;
			switch(u->key - 1) {
				case NONE:
					sprintf(str, "No Link");
					break;
				case ETHERNET_IP:
					sprintf(str, "Ethernet (IP):");
					break;
				case ETHERNET_ARP:
					sprintf(str, "Ethernet (ARP):");
					break;
					}
			pcnt = 100.0 * (Double)b/(Double)tb;
			print_line(str, u->num_packets, b, pcnt);
			}
		break;
		case TYPE_IP_PROTO:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];
			b = u->bytes;
			pcnt = 100.0 * (Double)b/(Double)tb;
			print_line(
				my_get_proto(
				  (u_short)u->key - 1, numbers_only),
				u->num_packets, b, pcnt);
			}
			break;
		case TYPE_PORT:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];
			b = u->bytes; pcnt = 100.0 * (Double)b/(Double)tb;
			print_line(
				my_get_port((u_short)u->key, numbers_only),
				u->num_packets, b, pcnt);
			}
			break;
		case TYPE_HOST:
		case TYPE_NET:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];
			b = u->bytes; pcnt = 100.0 * (Double)b/(Double)tb;
			i = u->key;
			addr = (struct in_addr *)&i;
			sprintf(str, "%s", inet_ntoa(*addr));
			print_line(str, u->num_packets, b, pcnt);
			}
			break;
		default:
		for (l=0; l<*count; l++) {
			u = (*hash)[l];
			b = u->bytes; pcnt = 100.0 * (Double)b/(Double)tb;
			printf("\t%-20X%-10qu%-10qu%-5.4f %%\n",
				u->key, u->num_packets, b, pcnt);
			}
			break;
		}
}

void add_packet_to_root(stat_root *roots, packet_data *pd, u_int types) {
	u_int tc, i, type;
	u_int key1, key2;

	tc=count_1bits(types);
	for (i=0; i<tc; i++) {
		type=roots[i].type;
		switch (type) {
			case TYPE_PORT:
			case TYPE_HOST:
			case TYPE_NET:
				/* Add src */
			key1 = get_data_from_packet(type,pd,1);
			if (key1)
				add_packet(type,
					&(roots[i].count),
					&(roots[i].hash),
					pd, key1);
				/* Add dest */
			key2 = get_data_from_packet(type,pd,0);
			if (key2 && key2 != key1)	/* Don't count duplicates */
				add_packet(type,
					&(roots[i].count),
					&(roots[i].hash),
					pd, key2);
				break;
			default:
			key1 = get_data_from_packet(type,pd,1);
			if (key1)
				add_packet(type,
					&(roots[i].count),
					&(roots[i].hash),
					pd, key1);
				break;
			}
		}
}

void show_results(stat_root *roots, u_int types) {
	int tc, i;
	u_quad_t tot;

	tc=count_1bits(types);
	tot = 0;
	for (i=0; i<roots[0].count; i++)
		tot += (roots[i].hash)[0]->bytes;
	for (i=0; i<tc; i++) {
		printf("%s Statistcs:\n", type_str[ffs(roots[i].type)-1] );
		compact_type(&(roots[i].count),	&(roots[i].hash) );
		qsort(
	roots[i].hash, roots[i].count, sizeof(void *), bytes_compare);
		show_totals(roots[i].type, &(roots[i].count),
				&(roots[i].hash), tot);
		}
}
