/*
 * Copyright (c) 2000 Paul Herman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "tcpprof.h"

#define KIL		(1024)
#define MEG		(1024*1024)
#define GIG		(1024*1024*1024)
#define Double		double		/* FP accuracy */
#define DEFAULT_INTERVAL	5
#define DEFAULT_FORMAT		"Time:%S\tn=%n\tavg=%a\tstddev=%d\tbps=%b\n"

enum {
	ACTION_INTERVAL, ACTION_BANDWIDTH, ACTION_ACCOUNT
};

/**
 **	GLOBALS - options to be set on the command line
 **/
char action			= ACTION_INTERVAL;
char *filterexpr = NULL;
char *outputformat = NULL;
Double	interval;
Double	relative_seconds	= 0;
Double	bandwidth		= 0;
char	gstr[BUF_SIZ];
int	get_tcp_flags		= 0;
Double	capture_seconds		= -1.0;

typedef struct statistics {
	int		count;
	Double		ts_begin;
	Double		ts_now;
	Double		sum;
	Double		sum2;
	Double		total_time;
	Double		total_time_over_bps;
	Double		total_bytes;
} statistics;

char *wr_bytes(Double b) {
        if (0) {
                snprintf(gstr, BUF_SIZ, "%12.0f Bytes", b); return gstr;  }
        if ( b > GIG )  {
                snprintf(gstr, BUF_SIZ, "%8.1f GB", b/GIG); return gstr;  }
        if ( b > MEG )  {
                snprintf(gstr, BUF_SIZ, "%8.1f MB", b/MEG); return gstr;  }
        if ( b > KIL )  {
                snprintf(gstr, BUF_SIZ, "%8.1f KB", b/KIL); return gstr;  }
        snprintf(gstr, BUF_SIZ, "%8.1f  B", b); return gstr;
}

void show_interval(statistics *s, Double dt) {
	char *f;
	Double avg, stddev, t_bar, delta_t;

	if (action != ACTION_INTERVAL) return;

	if (s->count == 0) {
		avg = 0;
		stddev = 0;
		}
	else {
		avg = s->sum/(Double)s->count;
		stddev = sqrt(s->sum2/(Double)s->count - avg*avg);
		}
	delta_t = (interval == -1)? (s->ts_now - s->ts_begin) : dt;
	t_bar = s->ts_begin + delta_t/2.0;
	f = outputformat;
	while (*f) {
		if (*f == '%') {
			switch (*(++f)) {
				case 'a':
					printf("%.2f", avg);
					break;
				case 'B':
					printf("%.2f", s->sum/delta_t);
					break;
				case 'b':
					printf("%.2f", s->sum/delta_t * 8.0);
					break;
				case 'n':
					printf("%d", s->count);
					break;
				case 'N':
					printf("%.0f", s->sum);
					break;
				case 'p':
					printf("%.2f", (Double)s->count/delta_t);
					break;
				case 'd':
					printf("%.2f", stddev);
					break;
				case 'S':
					printf("%d",
			(int) (t_bar + 0.5 - relative_seconds) );
					break;
				case 's':
					printf("%.6f", t_bar - relative_seconds);
					break;
				default:
					printf("%%%c", *f);
					break;
				}	/* switch */
			} /* if %  */
		else if (*f == '\\') {
			switch (*(++f)) {
			/* XXX: This could be written better */
				case 'a': printf("\a"); break;
				case 'b': printf("\b"); break;
				case 'f': printf("\f"); break;
				case 'n': printf("\n"); break;
				case 'r': printf("\r"); break;
				case 't': printf("\t"); break;
				case 'v': printf("\v"); break;
				default: printf("%c", *f); break;
				}
			}
		else printf("%c", *f);
		f++;
		}
}

void my_hook(packet_data *pd, void **args) {
	statistics	*sp, zero;
	Double		ts, x, ival, bps;

	sp = (statistics *) args[0];

	ival = interval;

	ts = 	(Double)pd->timestamp.tv_sec +
		(Double)pd->timestamp.tv_usec/1.0e6;
	if (sp->ts_begin == 0) {
			/* initialize stats */
		sp->ts_begin 	= ts;
		sp->count	= 0;
		sp->sum		= 0;
		sp->sum2	= 0;
		}
	if ( (ts - sp->ts_begin) > ival && interval != -1 ) {
			/* show interval and reset stats */
		show_interval(sp, ival);
		sp->ts_begin += ival;

		sp->total_time	+= ival;
		sp->total_bytes	+= sp->sum;
		bps = (sp->sum / ival) * 8.0;
		if (bps > bandwidth) sp->total_time_over_bps += ival;

		zero.count	= 0;
		zero.sum	= 0;
		zero.sum2	= 0;
		for (x = sp->ts_begin + ival; x<ts; x += ival) {
			zero.ts_begin = x - ival;
			sp->ts_begin += ival;
			sp->total_time	+= ival;
			show_interval(&zero, ival);
			}
		sp->count	= 0;
		sp->sum		= 0;
		sp->sum2	= 0;
		}
	sp->ts_now = ts;
	(sp->count)++;
	sp->sum += (Double)pd->packet_len;
	sp->sum2 += (Double)pd->packet_len * (Double)pd->packet_len;

}

/*
 * process_file() takes the output of tcpdump, saves packets, and displays
 * statistics
 */
void process_file(char *fname, u_int unused) {
	void		*argv[2];
	statistics	stats;
	Double		bps;

/* XXX: Want to add live tcpdumps option (no filename) with alarm()	*/
/* XXX: ...but that wouldn't allow multiple passes with filters...	*/

	stats.count	= 0;
	stats.ts_begin	= 0;
	stats.sum	= 0;
	stats.sum2	= 0;
	stats.total_bytes	= 0;
	stats.total_time	= 0;
	stats.total_time_over_bps = 0;

	argv[0] = (void *) &stats;

        get_tcpdump_data(fname, filterexpr, get_tcp_flags, capture_seconds, 
		my_hook, argv);

	switch (action) {
		case ACTION_INTERVAL:
			show_interval(&stats, (Double) interval);
			break;
		case ACTION_BANDWIDTH:
			printf("Time exceeding %.2f bps:\t%.2f sec.\t(%.2f%% of the time)\n",
				bandwidth, stats.total_time_over_bps,
				100*stats.total_time_over_bps/stats.total_time );
			break;
		case ACTION_ACCOUNT:
			if (stats.total_time != 0.0)
				bps = stats.total_bytes/stats.total_time;
			else bps = stats.total_bytes/interval;
			printf("Bytes/sec\t= %s\n", wr_bytes(bps));
			printf("Bytes/minute\t= %s\n", wr_bytes(bps*60));
			printf("Bytes/hour\t= %s\n", wr_bytes(bps*60*60));
			printf("Bytes/day\t= %s\n", wr_bytes(bps*60*60*24));
			printf("Bytes/month\t= %s\n", wr_bytes(bps*60*60*24*30));
			break;
		}
}

/* XXX: Want accounting mode -a for bytes per
 * 		second/minute/hour/month...
 * XXX: Want interval (=-1 or something) to signify entire length of sample
 */

#define USAGE 							\
"Usage: %s [-?halp] [-b bps] [-f filter expr] [-i interface] [-o output]
		[-R seconds] [-r filename] [-s seconds] [interval]
	-?, -h 		- display this help
	-a 		- accounting mode
	-b bps		- bandwidth mode, where bps is bits/second
	-f filter expr 	- filter expression to pass to tcpdump
	-i interface	- do live capture on [interface], not from file
	-l		- include linklayer in size calc (Ethernet only)
	-p		- non-promiscuous mode when doing live capture
	-o output 	- format for the output of stats (see below)
	-R seconds 	- display time relative to [seconds]
	-r filename 	- read data from [filename]
	-s seconds 	- capture only [seconds] long (-1 is infinite)

	interval	- time interval (seconds) for taking samples

OUTPUT FORMAT: (see manpage for a complete description)
	%%S		- time in SECONDS (rounded)
	%%n		- number of packets
	%%N		- number of bytes
	%%a		- average packet size
	%%b		- bits per second
"

int Usage(int r, char *prog) {
	fprintf(stderr, USAGE, my_basename(prog));
	return r;
}

void error(char *s) {
	fprintf(stderr, "error: %s\n", s);
	Usage(-1, "tcpstat"); exit(-1);
}

int main(int argc, char **argv) {
	char ch, *filename;

		/* Default promisc, like tcpdump(1) */
	get_tcp_flags = GET_TCPD_DO_LIVE_PROMISC;
	filename = NULL;
	my_safe_strcpy(&filterexpr, " ");
	my_safe_strcpy(&outputformat, DEFAULT_FORMAT);

	while ( (ch = getopt(argc, argv, "h?alpb:f:i:o:R:r:s:")) != -1) {
		switch(ch) {
			case 'h':
			case '?':
				return Usage(1, argv[0]);
				break;
			case 'a':
				action = ACTION_ACCOUNT;
				break;
			case 'l':
				get_tcp_flags |= GET_TCPD_COUNT_LINKSIZE;
				break;
			case 'p':
				get_tcp_flags &= ~GET_TCPD_DO_LIVE_PROMISC;
				break;
			case 'b':
				bandwidth = atof(optarg);
				if (bandwidth <= 0) return Usage(1, argv[0]);
				action = ACTION_BANDWIDTH;
				break;
			case 'f':
				my_safe_strcpy(&filterexpr, optarg);
				break;
			case 'i':
				get_tcp_flags |= GET_TCPD_DO_LIVE;
				my_safe_strcpy(&filename, optarg);
				break;
			case 'o':
				my_safe_strcpy(&outputformat, optarg);
				break;
			case 'R':
				relative_seconds = atof(optarg);
				if (relative_seconds < 0) return Usage(1, argv[0]);
				break;
			case 'r':
				get_tcp_flags &= ~GET_TCPD_DO_LIVE;
				my_safe_strcpy(&filename, optarg);
				break;
			case 's':
				capture_seconds = atof(optarg);
				if (capture_seconds <= 0 &&
					capture_seconds != -1)
					return Usage(1, argv[0]);
				break;
			default:
				break;
			}				
		}

	if ((argc - optind) < 1) interval = DEFAULT_INTERVAL;
	else {
		interval = atof(argv[optind]);
		if (interval <= 0 && interval != -1)
			return Usage(1, argv[0]);
		}

		/* Sanity checks */
	if ( ! (get_tcp_flags & (GET_TCPD_DO_LIVE|GET_TCPD_DO_LIVE_PROMISC)) )
		error("-p option not compatible with -r");

	if (filename == NULL) {
			/* Default read from interface */
		get_tcp_flags |= GET_TCPD_DO_LIVE;
		my_safe_strcpy(&filename, "auto");
		}
	process_file(filename, 0);
	if (filename) free(filename);

	return 0;
}
